#ifdef ATARI_ST
/*
 * The console driver for the Atari ST
 */
#define DEBDMP		/* dump system tables */
#define DEBOUT		/* debugging on console and/or printer */

#include "../h/const.h"
#include "../h/type.h"
#include "../h/signal.h"

#include "const.h"
#include "type.h"
#include "proc.h"
#include "tty.h"

#ifdef DEBOUT
PRIVATE int	STuseCon = 0;
PRIVATE int	STusePrt = 0;
#endif
#ifdef DEBDMP
PRIVATE int	STdoDump = 0;
#endif

/*===========================================================================*
 *				coninit					     *
 *===========================================================================*/
PUBLIC coninit()
{
  vduinit();
  kbdinit();
#ifdef DEBOUT
  STuseCon = 1;		/* debugging putc() goes (also) to console */
#endif
}

/*===========================================================================*
 *				operator				     *
 *===========================================================================*/
PUBLIC operator(pfx)
{
/* handle CTRL-ALT-PFX sequences */
  switch (pfx) {
#ifdef DEBDMP
  case 1:	/* PF1: print process table */
	p_dmp(); return;
  case 2:	/* PF2: print memory map */
	map_dmp(); return;
  case 3:	/* PF3: print regs user prog */
	reg_dmp(bill_ptr); return;
  case 6:	/* PF6: ON/OFF dump tables on panic */
	STdoDump ^= 1; return;
#endif
#ifdef DEBOUT
  case 4:	/* PF4: ON/OFF console debugging */
	STuseCon ^= 1; return;
  case 5:	/* PF5: ON/OFF printer debugging */
	STusePrt ^= 1; return;
#endif
  case 10:	/* PF10: issue SIGKILL */
	sigchar(&tty_struct[CONSOLE], SIGKILL); return;
  }
}

/*===========================================================================*
 *				dump					     *
 *===========================================================================*/
dump() {
#ifdef DEBDMP
  if (STdoDump == 0)
	return;
  printf("dump\n");
  reg_dmp(proc_ptr);
  p_dmp();
  map_dmp();
#endif
}

/*===========================================================================*
 *				putc					     *
 *===========================================================================*/
/* 
 * This procedure is used by the kernel version of printf().
 * The one in the library sends a message to FS, which is not
 * needed for printing within the kernel.
 */
PUBLIC putc(c)
{
#ifdef DEBOUT
/*
 * Use any combination of the printer or the console.
 */

  if (STuseCon) {
#endif
	vducursor(0);
	vduput(c);
	vducursor(1);
#ifdef DEBOUT
  }
  if (STusePrt)
	prtc(c);
#endif
}
#endif ATARI_ST
