#ifdef ATARI_ST
/* This file contains some dumping routines for debugging. */

#include "../h/const.h"
#include "../h/type.h"
#include "../h/callnr.h"
#include "../h/com.h"
#include "../h/error.h"
#include "const.h"
#include "type.h"
#include "glo.h"
#include "proc.h"

#undef	NEEDED

#define NSIZE 20
phys_bytes aout[NR_PROCS];	/* pointers to the program names */
char nbuff[NSIZE+1];
char *vargv;

/*===========================================================================*
 *				DEBUG routines here			     * 
 *===========================================================================*/
p_dmp()
{
/* Proc table dump */

  register struct proc *rp;
  char *np;
  vir_clicks base, limit, first, last;
  phys_bytes ltmp, dst;
  int index;
  extern phys_bytes umap();

  printf(
  "\nproc    pid     pc     sp  splow flag  user    sys  base limit recv   command\n");

  dst = umap(proc_addr(SYSTASK), D, (vir_bytes)nbuff, (vir_bytes)NSIZE);

  for (rp = &proc[0]; rp < &proc[NR_PROCS+NR_TASKS]; rp++)  {
	if (rp->p_flags & P_SLOT_FREE) continue;
	first = rp->p_map[T].mem_phys;
	last = rp->p_map[S].mem_phys + rp->p_map[S].mem_len;
	ltmp = ((long) first << CLICK_SHIFT) + 512L;
	base = (vir_clicks) (ltmp/1024L);
	ltmp = (((long) last << CLICK_SHIFT) + 512L);
	limit = (vir_clicks) (ltmp/1024L);
	prname((int)(rp - proc));
	printf(" %4d %6X %6X %6X %4x %5D %6D  %3dK %3dK  ",
		rp->p_pid, (long)rp->p_pcpsw.pc,
		(long)rp->p_sp, (long)rp->p_splow,
		rp->p_flags, rp->user_time, rp->sys_time,
		base, limit);
	if (rp->p_flags == 0)
		printf("      ");
	else
		prname(NR_TASKS + rp->p_getfrom);

	/* Fetch the command string from the user process. */
	index = (int)(rp - proc) - NR_TASKS;
	if (index > LOW_USER && aout[index] != 0) {
		phys_copy(aout[index], dst, (long) NSIZE);
		nbuff[NSIZE] = 0;
		for (np = &nbuff[0]; np < &nbuff[NSIZE]; np++)
			if (*np <= ' ' || *np >= 0177) *np = 0;
		if (index == LOW_USER + 1)
			printf("/bin/sh");
		else
			printf("%s", nbuff);
	}
	printf("\n");
  }
#ifdef NEEDED
  printf("\nproc    callq sendlk mesbuf from nready pend\n");
  for (rp = &proc[0]; rp < &proc[NR_PROCS+NR_TASKS]; rp++)  {
	if (rp->p_flags & P_SLOT_FREE) continue;
	prname((int)(rp - proc));
	printf(" %6X %6X %6X %4d %6X %4d\n",
		rp->p_callerq,
		rp->p_sendlink,
		rp->p_messbuf,
		rp->p_getfrom,
		rp->p_nextready,
		rp->p_pending
	);
  }
  for (index = 0; index < NQ; index++) {
	printf("q=%d: ", index);
	for (rp = rdy_head[index]; rp != NIL_PROC; rp = rp->p_nextready)
		printf(" %d", (int)(rp-proc));
	printf("\n");
  }
#endif NEEDED
  printf("\n");
}



map_dmp()
{
  register struct proc *rp;
  vir_clicks base, limit, first, last;
  phys_bytes ltmp;

  printf("\nPROC   --TEXT---  --DATA---  --STACK-- SHADOW FLIP P BASE SIZE\n");
  for (rp = &proc[NR_TASKS]; rp < &proc[NR_TASKS+NR_PROCS]; rp++)  {
	if (rp->p_flags & P_SLOT_FREE) continue;
	first = rp->p_map[T].mem_phys;
	last = rp->p_map[S].mem_phys + rp->p_map[S].mem_len;
	ltmp = ((long) first << CLICK_SHIFT) + 512L;
	base = (vir_clicks) (ltmp/1024L);
	ltmp = (((long) (last-first) << CLICK_SHIFT) + 512L);
	limit = (vir_clicks) (ltmp/1024L);
	prname((int)(rp-proc));
	printf(" %4x %4x  %4x %4x  %4x %4x   %4x %4d %d %3dK %3dK\n", 
	    rp->p_map[T].mem_phys, rp->p_map[T].mem_len,
	    rp->p_map[D].mem_phys, rp->p_map[D].mem_len,
	    rp->p_map[S].mem_phys, rp->p_map[S].mem_len,
	    rp->p_shadow, rp->p_nflips, rp->p_physio, base, limit);
  }
}


char *nayme[]= {"PRINTR", "TTY   ", "WINCHE", "FLOPPY", "RAMDSK", "CLOCK ", 
		"SYS   ", "HARDWR", "MM    ", "FS    ", "INIT  "};
prname(i)
int i;
{
  if (i == ANY+NR_TASKS)
	printf("ANY   ");
  else if (i >= 0 && i <= NR_TASKS+2)
	printf("%s",nayme[i]);
  else
	printf("%4d  ", i-NR_TASKS);
}

set_name(proc_nr, ptr)
int proc_nr;
char *ptr;
{
/* When an EXEC call is done, the kernel is told about the stack pointer.
 * It uses the stack pointer to find the command line, for dumping
 * purposes.
 */

  extern phys_bytes umap();
  phys_bytes src, dst;

  if (ptr == (char *) 0) {
	aout[proc_nr] = (phys_bytes) 0;
	return;
  }

  src = umap(
	proc_addr(proc_nr),
	D,
	(vir_bytes)(ptr + sizeof(char *)),
	(vir_bytes)sizeof(char *)
  );
  if (src == 0)
	return;
  dst = umap(
	proc_addr(SYSTASK),
	D,
	(vir_bytes)&vargv,
	(vir_bytes)sizeof(char *)
  );
  phys_copy(src, dst, (phys_bytes)sizeof(char *));
  aout[proc_nr] = umap(
	proc_addr(proc_nr),
	D,
	(vir_bytes)vargv,
	(vir_bytes)NSIZE
  );
}

reg_dmp(rp, dummy)
register struct proc *rp;
{
	register i;
	static char *regs[NR_REGS] = {
		"d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7",
		"a0", "a1", "a2", "a3", "a4", "a5", "a6"
	};
	
	printf("reg = %08X, ", rp);
	printf("ksp = %08X\n", (long)&dummy);
	printf(" pc = %08X, ", rp->p_pcpsw.pc);
	printf(" sr =     %04x, ", rp->p_pcpsw.psw);
	printf("trp =       %2x\n", rp->p_trap);
	for (i = 0; i < NR_REGS; i++)
		printf("%3s = %08X%s",
			regs[i],
			rp->p_reg[i],
			(i&3) == 3 ? "\n" : ", "
		);
	printf(" a7 = %08X\n", rp->p_sp);
	mem_dmp((char *)(((long)rp->p_pcpsw.pc & ~31L) - 96), 128);
	mem_dmp((char *)(((long)rp->p_sp & ~31L) - 32), 256);
}

mem_dmp(adr, len)
char *adr;
{
	register i;
	register long *p;

	for (i = 0, p = (long *)adr; i < len; i += 4) {
		if ((i & 31) == 0)
			printf("\n%X:", p);
		printf(" %8X", *p++);
	}
	printf("\n");
}
#endif
