/* replace undef by define */
#undef	 DEBUG		/* check assertions */
#undef	 SLOWDEBUG	/* some extra test loops (requires DEBUG) */

#ifdef DEBUG
#define	ASSERT(b)	if (!(b)) assert_failed();
#else
#define	ASSERT(b)	/* empty */
#endif

#ifdef i8088
#define	ptrint		int
#endif
#ifdef ATARI_ST
#define	ptrint		long
#endif

#define BRKSIZE		1024
#define	PTRSIZE		sizeof(char *)
#define Align(x,a)	(((x) + (a - 1)) & ~(a - 1))
#define NextSlot(p)	(* (char **) ((p) - PTRSIZE))
#define NextFree(p)	(* (char **) (p))

/*
 * A short explanation of the data structure and algorithms.
 * An area returned by malloc() is called a slot. Each slot
 * contains the number of bytes requested, but preceeded by
 * an extra pointer to the next the slot in memory.
 * 'bottom' and 'top' point to the first/last slot.
 * More memory is asked for using brk() and appended to top.
 * The list of free slots is maintained to keep malloc() fast.
 * 'empty' points the the first free slot. Free slots are
 * linked together by a pointer at the start of the
 * user visable part, so just after the next-slot pointer.
 * Free slots are merged together by free().
 */

extern char *sbrk(), *brk();
static char *bottom, *top, *empty;

static grow(len)
unsigned len;
{
  register char *p;

  ASSERT(NextSlot(top) == 0);
  p = (char *) Align((ptrint)top + len, BRKSIZE);
  if (p < top || brk(p) != 0)
	return(0);
  NextSlot(top) = p;
  NextSlot(p) = 0;
  free(top);
  top = p;
  return(1);
}

char *malloc(size)
unsigned size;
{
  register char *prev, *p, *next, *new;
  register unsigned len, ntries;

  if (size == 0)
	size = PTRSIZE;		/* avoid slots less that 2*PTRSIZE */
  for (ntries = 0; ntries < 2; ntries++) {
	len = Align(size, PTRSIZE) + PTRSIZE;
	if (bottom == 0) {
		p = sbrk(2 * PTRSIZE);
		p = (char *) Align((ptrint)p, PTRSIZE);
		p += PTRSIZE;
		top = bottom = p;
		NextSlot(p) = 0;
	}
#ifdef SLOWDEBUG
	for (p = bottom; (next = NextSlot(p)) != 0; p = next)
		ASSERT(next > p);
	ASSERT(p == top);
#endif
	for (prev = 0, p = empty; p != 0; prev = p, p = NextFree(p)) {
		next = NextSlot(p);
		new = p + len;
		if (new > next)
			continue;		/* too small */
		if (new + PTRSIZE < next) {	/* too big, so split */
			/* + PTRSIZE avoids tiny slots on free list */
			NextSlot(new) = next;
			NextSlot(p) = new;
			NextFree(new) = NextFree(p);
			NextFree(p) = new;
		}
		if (prev)
			NextFree(prev) = NextFree(p);
		else
			empty = NextFree(p);
		return(p);
	}
	if (grow(len) == 0)
		break;
  }
  ASSERT(ntries != 2);
  return(0);
}

char *realloc(old, size)
char *old;
unsigned size;
{
  register char *prev, *p, *next, *new;
  register unsigned len, n;

  len = Align(size, PTRSIZE) + PTRSIZE;
  next = NextSlot(old);
  n = (int)(next - old);			/* old length */
  /*
   * extend old if there is any free space just behind it
   */
  for (prev = 0, p = empty; p != 0; prev = p, p = NextFree(p)) {
	if (p > next)
		break;
	if (p == next) {	/* 'next' is a free slot: merge */
		NextSlot(old) = NextSlot(p);
		if (prev)
			NextFree(prev) = NextFree(p);
		else
			empty = NextFree(p);
		next = NextSlot(old);
		break;
	}
  }
  new = old + len;
  /*
   * Can we use the old, possibly extended slot?
   */
  if (new <= next) {				/* it does fit */
	if (new + PTRSIZE < next) {		/* too big, so split */
		/* + PTRSIZE avoids tiny slots on free list */
		NextSlot(new) = next;
		NextSlot(old) = new;
		free(new);
	}
	return(old);
  }
  if ((new = malloc(size)) == 0)		/* it didn't fit */
	return(0);
  bcopy(old, new, n);				/* n < size */
  free(old);
  return(new);
}

static bcopy(old, new, n)
register char *old, *new;
register n;
{
  while (n--)
	*new++ = *old++;
}

char *calloc(n, size)
unsigned n, size;
{
  register char *p, *cp;

  n *= size;
  cp = malloc(n);
  if (cp == (char *)0)
	return ((char *)0);
  for (p = cp; n-- != 0; )
	*p++ = '\0';
  return(cp);
}

free(p)
char *p;
{
  register char *prev, *next;

  ASSERT(NextSlot(p) > p);
  for (prev = 0, next = empty; next != 0; prev = next, next = NextFree(next))
	if (p < next)
		break;
  NextFree(p) = next;
  if (prev)
	NextFree(prev) = p;
  else
	empty = p;
  if (next) {
	ASSERT(NextSlot(p) <= next);
	if (NextSlot(p) == next) {		/* merge p and next */
		NextSlot(p) = NextSlot(next);
		NextFree(p) = NextFree(next);
	}
  }
  if (prev) {
	ASSERT(NextSlot(prev) <= p);
	if (NextSlot(prev) == p) {		/* merge prev and p */
		NextSlot(prev) = NextSlot(p);
		NextFree(prev) = NextFree(p);
	}
  }
}

#ifdef DEBUG
static assert_failed()
{
	write(2, "assert failed in lib/malloc.c\n", 30);
	abort();
}
#endif
